import os, yaml, itertools
from collections import defaultdict
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.patches as patches
from matplotlib.lines import Line2D
import colorcet as cc

doAll = False


# Function to extract statistics from a line
def extract_stat(line):
    prefix, value = line.split(":")
    clean_prefix = " ".join(prefix.split())  # Remove extra spaces
    err = 0
    value = value.replace("+/-", "±")
    if "±" in value:
        err = float(value.strip().split("±")[1].rstrip("%").strip())
    return clean_prefix, (float(value.strip().split()[0].rstrip("%")), err)


def compute_stat(line):
    n_str, result_str = line.split("\t", 1)
    result_str = result_str.split("\t")[0]
    n = float(n_str)
    result = float(result_str)
    correct=round(n*(result/100.0))
    p=correct/n
    err=((p*(1-p)/(n-1))**0.5)*100.0
    return (float(result), err)


# Directories containing the result and evaluation files
resultsDirectory = "results"
evaluationDirectory = "evaluation"

models = set()
perplexity_stats = defaultdict(lambda: defaultdict(dict))
kl_divergence_stats = defaultdict(lambda: defaultdict(dict))
token_probability_stats = defaultdict(lambda: defaultdict(dict))
arc_easy_stats = defaultdict(dict)
arc_challenge_stats = defaultdict(dict)
mmlu_stats = defaultdict(dict)
winogrande_stats = defaultdict(dict)
eval_stats = defaultdict(dict)

os.makedirs("out", exist_ok=True)

# Iterate over each file in the directories
for entry in itertools.chain(os.scandir(resultsDirectory), os.scandir(evaluationDirectory)):
    if not entry.name.endswith(".txt"):
        continue
    if "Qwen2.5" in entry.name:
        continue
    #if "405B" not in entry.name:
    #    continue
    if not doAll:
        if "i1" in entry.name:
            if "IQ1_S" in entry.name:
                continue
            if "IQ1_M" in entry.name:
                continue
        else:
            if "IQ3_XS" in entry.name:
                continue
            if "IQ3_S" in entry.name:
                continue
            if "IQ3_M" in entry.name:
                continue
    print(f"Processing file: {entry.path}")
    
    # Extract the base name and quant from the file name
    if entry.path.startswith("evaluation"):
        splits = entry.name.rsplit('.', 3)
    else:
        splits = entry.name.rsplit('.', 2)
    model = splits[0]
    quant = splits[1]
    models.add(model)
    
    with open(entry.path, encoding="utf-8") as file:
        print(entry.path)
        if entry.name.endswith(".arc-easy.txt"):
            for line in file:
                if line.startswith("570\t"):
                    arc_easy_stats[model][quant] = compute_stat(line)
        elif entry.name.endswith(".arc-challenge.txt"):
            for line in file:
                if line.startswith("299\t"):
                    arc_challenge_stats[model][quant] = compute_stat(line)
        elif entry.name.endswith(".mmlu.txt"):
            for line in file:
                if line.startswith("1548\t"):
                    mmlu_stats[model][quant] = compute_stat(line)
        elif entry.name.endswith(".winogrande.txt"):
            for line in file:
                if line.startswith("1267\t"):
                    winogrande_stats[model][quant] = compute_stat(line)
        else:
            for line in file:
                if ":" in line and not "multiple_choice_score" in line:
                    key, value = extract_stat(line)
                    if "PPL" in key:
                        perplexity_stats[key][model][quant] = value
                    elif "KLD" in key:
                        kl_divergence_stats[key][model][quant] = value
                    elif "Δp" in key or "Same top p" in key:
                        token_probability_stats[key][model][quant] = value


for model in models:
    print(model)
    for quant in arc_easy_stats[model].keys():
        #print(quant)
        normArcEasy = (arc_easy_stats[model][quant][0]/arc_easy_stats[model]["SOURCE"][0]) * 570
        normArcChallenge = (arc_challenge_stats[model][quant][0]/arc_challenge_stats[model]["SOURCE"][0]) * 299
        normMMLU = (mmlu_stats[model][quant][0]/mmlu_stats[model]["SOURCE"][0]) * 1548
        normWinogrande = (winogrande_stats[model][quant][0]/winogrande_stats[model]["SOURCE"][0]) * 1267
        wAvgEval = (normArcEasy + normArcChallenge + normMMLU + normWinogrande)/3684.0
        
        normArcEasyErr = (arc_easy_stats[model][quant][1]/arc_easy_stats[model]["SOURCE"][0]) * 570
        normArcChallengeErr = (arc_challenge_stats[model][quant][1]/arc_challenge_stats[model]["SOURCE"][0]) * 299
        normMMLUErr = (mmlu_stats[model][quant][1]/winogrande_stats[model]["SOURCE"][0]) * 1548
        normWinograndeErr = (winogrande_stats[model][quant][1]/winogrande_stats[model]["SOURCE"][0]) * 1267
        wAvgEvalErr = (normArcEasyErr + normArcChallengeErr + normMMLUErr + normWinograndeErr)/3684.0
        
        eval_stats[model][quant] = (wAvgEval*100.0, (wAvgEvalErr/wAvgEval)*100.0)

def plot_statistics(model, stats, title, ylabel, transform = lambda x: x, errTransform = lambda x: x, scale = lambda x: x):
    plt.figure(figsize=(30 if doAll else 28, 10))
    colors = cc.glasbey  # Use the glasbey colormap with at least 50 distinct colors
    color_map = {}
    marker_map = {}

    # Load the YAML file
    yaml_file_path = f'size/{model}.yaml'
    with open(yaml_file_path, 'r') as file:
        size_data = yaml.load(file, Loader=yaml.FullLoader)
    
    minValue = 1000
    maxValue = -1000
    
    rawValues = []
    
    for quant, data in sorted(stats.items(), key=lambda x: size_data.get(x[0], 0) + 100000 if "i1" in x[0] else size_data.get(x[0], 0)):
        quants = []
        shortLabels = []

        quants.append(size_data[quant] / size_data["SOURCE"] * 100)  # Use size data for y-axis
        rawValue = (transform(data[0]))
        rawErr = (errTransform(data[1]))
        rawValues.append((rawValue, rawErr))
        value = (scale(rawValue))
        shortQuant = quant.replace("i1-", "i")
        
        if minValue > value:
            minValue = value
        if maxValue < value:
            maxValue = value
        
        relativeSize = (size_data[quant] / size_data["SOURCE"])*100.0
        sizeGB = size_data[quant]/(1000*1000*1000)
        lenIntGB = len(str(int(sizeGB)))
        shortLabel = f'{shortQuant}\n{relativeSize:.2f}%\n{sizeGB:.{4-lenIntGB}f}G'
        if quant not in color_map:
            color_map[quant] = colors[len(color_map) % len(colors)]
        #marker = 'o' if 'i1' in quant else '^'
        #marker_map[quant] = marker
        plt.bar([shortLabel], [value], color=color_map[quant], zorder=5)

    # Add the following lines before the for loop that iterates over the patches
    ax = plt.gca()  # Get the current Axes object
    patches = ax.get_children()
    
    # Then, replace the following lines with the modified code:
    # Add labels on top of bars
    for (i, rect), value in zip(enumerate(patches), rawValues):
        if hasattr(rect, 'get_height'):
            height = rect.get_height()
            plt.text(rect.get_x() + rect.get_width() / 2.0,
                    height,
                    str(f"{value[0]:.{5-len(str(int(value[0])))}f}\n±{value[1]:.{5-len(str(int(value[1])))}f}"),
                    ha='center',
                    va='bottom')

    plt.title(title)
    plt.xlabel('Quant (i=imatrix) | Size relative to base [%] | Size [GB]')
    plt.ylabel(ylabel)
    plt.ylim(max(minValue-(maxValue-minValue)/20.0, 0), maxValue+(maxValue-minValue)/20.0)

    #plt.gca().yaxis.set_major_formatter(ticker.ScalarFormatter())
    #plt.gca().yaxis.set_minor_formatter(ticker.ScalarFormatter())

    #plt.yscale("symlog")
    plt.tight_layout()
    plt.margins(0.01, True)
    plt.savefig(f'out/{title.replace(":", " -")}.png')
    plt.close()
    #plt.show()

for model in models:
    plot_statistics(model, perplexity_stats["Mean PPL(Q)/PPL(base)"][model], f'{model}: Perplexity', 'Mean PPL(quant)/PPL(base) [%]', lambda x: x*100, lambda x: x*100)
    plot_statistics(model, kl_divergence_stats["Mean KLD"][model], f'{model}: KL Divergence', 'Mean KLD', lambda x: x)
    plot_statistics(model, token_probability_stats["Same top p"][model], f'{model}: Probability of quant generating the same token', 'Same top p [%]', lambda x: x)
    plot_statistics(model, token_probability_stats["Mean Δp"][model], f'{model}: Correct token probability', 'Correct token probability relative to base [%]', lambda x: x+100)
    #plot_statistics(model, arc_easy_stats[model], f'{model}: ARC Easy 0-shot', 'ARC Score', lambda x: x)
    #plot_statistics(model, arc_challenge_stats[model], f'{model}: ARC Challenge 0-shot', 'ARC Score', lambda x: x)
    #plot_statistics(model, mmlu_stats[model], f'{model}: MMLU 0-shot', 'MMLU Score', lambda x: x)
    #plot_statistics(model, winogrande_stats[model], f'{model}: Winogrande 0-shot', 'Winogrande Score', lambda x: x)
    plot_statistics(model, eval_stats[model], f'{model}: Eval (ARC, MMLU, Winogrande)', 'Eval score relative to base [%]', lambda x: x)
